/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "kernel.h"
#include "frmmain.h"
#include "frmview.h"
#include "glView.h"
#include "kernel_grid.h"
#include "frmsequence.h"
#include <qcursor.h>
#include <qpopupmenu.h>
#include <qinputdialog.h>
#include <rgCSG.h>

// Description: Draws the mouse adding/building triangle.
void hlp_draw_build_triangle( Model &mdl);
void hlp_draw_build_bone( const Model &mdl);

void vfn_initializeGL( GLView *g)
{
	glClearColor( 1, 1, 1, 1);
	//glEnable( GL_DEPTH_TEST);
	glClearDepth( -1);
	glDepthFunc( GL_GREATER);

	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	g->vmat.set_transformation_order( RGGL_VIEW_MATRIX_TRS);
	g->vmat.update();

	ViewData &vd = *((ViewData *)(g->edata));

	glFrontFace( GL_CW);
	glCullFace( GL_BACK);

	glEnable( GL_NORMALIZE);

	vd.light[0].set_pos( 0, 0, 1, 0);
	vd.light[0].set_amb( 0.2, 0.2, 0.2, 0);
	vd.light[0].set_dif( 1, 1, 1, 1);
	vd.light[0].set_spec( 1, 1, 1, 1);

	glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glEnable( GL_COLOR_MATERIAL);
}

void vfn_destructorGL( GLView *g)
{
	if( g->edata) delete ((ViewData *)(g->edata));
}

void hlp_show_vertex_selection_box( ViewData &vd)
{
	if( !(vd.show_vertex_sel_box)) return;
	rgMatrix m;
	switch( krn_get_edit_mode()) {
	case KED_MODE_ROTATE:
	case KED_MODE_SCALE:
		m = vd.frm->mdl->tool_matrix;
		break;
	}
	vd.frm->mdl->sel_box.render( m);
}

void vfn_paintGL( GLView *g)
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	loadRGMatrix( g->vmat.camM);

	ViewData &vd = *((ViewData *)(g->edata));
	sk3DRomModel &mdl = vd.frm->mdl->mdl;

	glEnable( GL_DEPTH_TEST);
	if( vd.edit_window && vd.use_wireframe) {
		glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
		// Draw back faces.
		glCullFace( GL_FRONT);
		glEnable( GL_CULL_FACE);
		glLineStipple( 2, 0xCCCC);
		glEnable( GL_LINE_STIPPLE);
		glColor3f( 0.6f, 0.6f, 0.0f);
		vd.frm->mdl->mdl.render_faces( false, false, 0, 0, true, true);
		glDisable( GL_LINE_STIPPLE);

		// Draw the bones
		glColor3f( 0, 0, 0);
		glLineWidth( 5);
		glBegin( GL_LINES);
		vd.frm->mdl->mdl.render_skel_frame();
		glEnd();
		glLineWidth( 1);

	} else glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);

	// Draw front faces.
	if( vd.use_wireframe || (vd.use_texture && vd.edit_window)) {
		glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
		glDisable( GL_DEPTH_TEST);
	} else {
		glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
		glEnable( GL_DEPTH_TEST);
	}
	glCullFace( GL_BACK);
	glEnable( GL_CULL_FACE);
	if( !(vd.edit_window) && (vd.use_lighting)) {
		/* FIXME : This line is temporary. */ vd.light[0].enable();
		glEnable( GL_LIGHTING);
	} else {
		glDisable( GL_LIGHTING);
	}
	if( (vd.use_texture && vd.edit_window)) {
		glEnable( GL_DEPTH_TEST);
		glEnable( GL_POLYGON_OFFSET_LINE);
		glPolygonOffset( 1, 1);
		vd.frm->mdl->mdl.render_faces( false, false, 0, 0, true, true);
		glDisable( GL_POLYGON_OFFSET_LINE);
		glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
		glEnable( GL_DEPTH_TEST);
	}
	if( vd.use_texture) {
		glEnable( GL_TEXTURE_2D);
		glColor3f( 1, 1, 1);
	} else glColor3f( 0, 0, 1);
	vd.frm->mdl->mdl.render_faces( vd.use_lighting, vd.use_texture, 0, 0,
		false, !(vd.use_texture));
	if( !(vd.edit_window)) {
		glEnable( GL_BLEND);
		if( vd.frm->mdl->frmSeq->m_playing) {
			vd.frm->mdl->mdl.render_bsp_faces(
				vd.frm->mdl->frmSeq->m_dc,
				g->vmat.icamM * rgVector( 0, 0, 4096.0f),
				vd.use_lighting, vd.use_texture, 0, 0);
		} else {
			vd.frm->mdl->mdl.render_bsp_faces(
				g->vmat.icamM * rgVector( 0, 0, 4096.0f),
				vd.use_lighting, vd.use_texture, -1);
		}
		glDisable( GL_BLEND);
	}
	glDisable( GL_TEXTURE_2D);


	// Draw the grid (if enabled)
	glDisable( GL_DEPTH_TEST);
	if( vd.edit_window && vd.show_grid) {
		glEnable( GL_BLEND);
		rgMatrix &m = g->vmat.icamM;
		draw_axis_aligned_grid( g, vd.small_grid_res, m, 8.0f, 0.125f,
			rgVector( 0.5, 0, 0),
			rgVector( 0.5, 0, 0),
			rgVector( 1, 0, 0));

		draw_axis_aligned_grid( g, vd.medium_grid_res, m, 2.0f, 0.25f,
			rgVector( 0.25, 0, 0),
			rgVector( 0.25, 0, 0),
			rgVector( 1, 0, 0));

		draw_axis_aligned_grid( g, vd.large_grid_res, m, 0.0f, 0.5f,
			rgVector( 0, 0, 0),
			rgVector( 0, 0, 0),
			rgVector( 1, 0, 0));
		glDisable( GL_BLEND);
	}

	glEnable( GL_DEPTH_TEST);
	// Show the current build state of the triangle (if in
	// triangle adding mode).
	if( KED_MODE_ADD_TRI == krn_get_edit_mode())
		hlp_draw_build_triangle( *(vd.frm->mdl));

	// If a bone is being added, then show the current build state.
	if( (KED_MODE_ADD_BONE == krn_get_edit_mode()))
		hlp_draw_build_bone( *(vd.frm->mdl));

	// Draw the model vertices.
	glEnable( GL_POLYGON_OFFSET_POINT);
	glPolygonOffset( 1, 1);
	if( vd.edit_window) {
		glPointSize( 5);
		glColor3f( 0, 0, 0);
		glBegin( GL_POINTS);
		vd.frm->mdl->mdl.render_group_verts();
		glEnd();
	}
	glDisable( GL_POLYGON_OFFSET_POINT);
	glDisable( GL_DEPTH_TEST);

	// Show the selected triangle
	if(	(KED_MODE_SEL_TRI == krn_get_edit_mode()) &&
		(vd.frm->mdl->sel_triangle >= 0)) {
		const int i = vd.frm->mdl->sel_triangle;
		const rgVector v[] = {
			mdl.get_tv_vertex( mdl.f[i].vid[0]),
			mdl.get_tv_vertex( mdl.f[i].vid[1]),
			mdl.get_tv_vertex( mdl.f[i].vid[2])};
		glLineWidth( 5);
		glCullFace( GL_FRONT);
		glBegin( GL_TRIANGLES);
		glColor3f( 1, 0, 0);
		glVertex( v[0]);
		glVertex( v[1]);
		glVertex( v[2]);
		glEnd();

		glCullFace( GL_BACK);
		glBegin( GL_TRIANGLES);
		glColor3f( 0, 1, 1);
		glVertex( v[0]);
		glVertex( v[1]);
		glVertex( v[2]);
		glEnd();
		glLineWidth( 1);

		glPointSize( 10);
		glBegin( GL_POINTS);
		glColor3f( 1, 0, 0); glVertex( v[0]);
		glColor3f( 0, 1, 0); glVertex( v[1]);
		glColor3f( 0, 0, 1); glVertex( v[2]);
		glEnd();
	}
	
	// Show the selected point
	if( vd.frm->mdl->sel_vertex >= 0) {
		glPointSize( 10);
		glBegin( GL_POINTS);
		glColor3f( 0, 1, 1);
		glVertex( mdl.get_tv_vertex( vd.frm->mdl->sel_vertex));
		glEnd();

	} else if( vd.frm->mdl->sel_verticies_count > 0) {
		glBegin( GL_POINTS);
		glColor3f( 1, 0, 1);
		mdl.render_verts(
			vd.frm->mdl->sel_verticies,
			vd.frm->mdl->sel_verticies_count);
		glEnd();
	}

	//glColor3f( 1, 0, 0);
	//glPointSize( 10);
	//glBegin( GL_POINTS);
	//glVertex( mdl.skel.m_lock_cs.p);
	//glEnd();

	// Show the vertex selection box.
	switch( krn_get_edit_mode()) {
	case KED_MODE_SEL_VERT:
	case KED_MODE_ROTATE:
	case KED_MODE_SCALE:
		hlp_show_vertex_selection_box( vd);
		break;
	}

	// Draw the selected bone.
	if( vd.edit_window) {
	switch( krn_get_edit_mode()) {
	case KED_MODE_SEL_VERT:
	case KED_MODE_SEL_BONE:
		if(	(vd.frm->mdl->do_bone_bind) ||
			(vd.frm->mdl->do_bone_mount)) {
			glLineWidth( 5);
			glBegin( GL_LINES);
			if( vd.frm->mdl->do_bone_bind) glColor3f( 0, 1, 0);
			else glColor3f( 1, 0, 0);
			glVertex( vd.frm->mdl->bone_build_v[2]);
			glVertex( vd.frm->mdl->bone_build_v[3]);
			glEnd();
			glLineWidth( 1);
		}

		if(	(vd.frm->mdl->sel_bone >= 0) &&
			(!(vd.frm->mdl->do_bone_bind || vd.frm->mdl->do_bone_mount) ||
			(vd.frm->mdl->sel_bone !=
				vd.frm->mdl->bone_bind_slave))) {
			rgVector ja = vd.frm->mdl->bone_build_v[0];
			rgVector jb = vd.frm->mdl->bone_build_v[1];
			glLineWidth( 5);
			glBegin( GL_LINES);
			glColor3f( 1, 0.5, 0);
			glVertex( ja);
			glVertex( jb);
			glEnd();
			glLineWidth( 1);

			if( mdl.skel.b[ vd.frm->mdl->sel_bone].pj > 0)
				rgSwap( ja, jb);
			glPointSize( 10);
			glBegin( GL_POINTS);
			glColor3f( 1, 0.5, 0);
			glVertex( ja);
			glColor3f( 0, 0, 1);
			glVertex( jb);
			glEnd();
		}
		break;
	}
	}

	glEnable( GL_DEPTH_TEST);
	if( vd.use_wireframe) {
		glDisable( GL_CULL_FACE);
		glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
		glLineWidth( 1);
	} else {
		glEnable( GL_CULL_FACE);
		glPolygonMode( GL_FRONT, GL_FILL);
	}
	mdl.m_csg->draw_all_brushes();

	if(	(mdl.m_brush_cursor) &&
		((KED_MODE_SEL_BRUSH == krn_get_edit_mode()) ||
		(KED_MODE_ADD_BRUSH == krn_get_edit_mode()) ||
		(KED_MODE_ROTATE_BRUSH == krn_get_edit_mode()) ||
		(KED_MODE_SCALE_BRUSH == krn_get_edit_mode()) ||
		(KED_MODE_CAMERA_TRAN == krn_get_edit_mode()))) {

		glPointSize( 5);
		glLineWidth( 1);
		glDisable( GL_DEPTH_TEST);
		glDisable( GL_CULL_FACE);
		mdl.m_brush_cursor->draw_cursor( rgVector( 1, 0, 0),
			rgVector( 0, 0, 1));
		glEnable( GL_CULL_FACE);
	}
}

void vfn_resizeGL( int w, int h, GLView *g)
{
	if( (w <= 0) || (h <= 0)) return;
	glViewport( 0, 0, w, h);

        if( !(g->edata)) return;
        ViewData &vd = *((ViewData *)(g->edata));
	vd.ortho_width = w;
	vd.ortho_height = h;

        glMatrixMode( GL_PROJECTION);
        glLoadIdentity();
        float haspect = 1.0f;
        float vaspect = 1.0f;
        //if( w >= h) haspect = float( w) / float (h);
        //else vaspect = float( h) / float( w);

	//if( vd.edit_window) {
	//	glClearDepth( -1);
	//	glDepthFunc( GL_GREATER);
		glOrtho( (vd.ortho_width * 0.5f) * -haspect,
			(vd.ortho_width * 0.5f) * haspect,
			(vd.ortho_height * 0.5f) * -vaspect,
			(vd.ortho_height * 0.5f) * vaspect,
			65535, -65535);
	//} else {
	//	glClearDepth( 1);
	//        glDepthFunc( GL_LESS);
	//	gluPerspective( 90.0f, 1.0f, 4.0f, 4096.0f);
	//}
	glMatrixMode( GL_MODELVIEW);
}

void hlp_do_vert_sel_mouse_down( QMouseEvent *e, GLView *g, ViewData &vd)
{
	Model &m = *(vd.frm->mdl);
	sk3DRomModel &mdl = vd.frm->mdl->mdl;

	// If the vertex selection box is visible and the user clicked within the
	// selection box, then the user is moving the selection box.
	if( vd.show_vertex_sel_box) {
		rgVector v( e->x(), e->y(), 0);
		rggl_ortho_window_to_world( g->width(), g->height(),
			&v, 1, &(g->vmat.icamM));
		if( m.sel_box.mouse_in_box( v)) {
			vd.do_sel_box_move = true;
			m.sel_box.set_mouse_down( krn_snap_to_grid( vd, v));
			return;
		}
	}

	m.sel_box.set_offset( rgVector());
	m.clear_verticies_selection();
	vd.do_sel_box_move = false;
	vd.show_vertex_sel_box = false;
	vd.do_vertex_move = false;
	if( m.sel_vertex >= 0) {
		rgVector v( e->x(), e->y(), 0);
		rggl_ortho_window_to_world( g->width(), g->height(),
			&v, 1, &(g->vmat.icamM));
		const rgVector u( rgMakePlanar(
			mdl.get_tv_vertex( m.sel_vertex), g->view_normal()));
		v = rgMakePlanar( v, g->view_normal());
		if( (v - u).length() < (vd.world_pixel_length * 16.0f)) {
			vd.do_vertex_move = true;
			g->setCursor( QCursor( Qt::CrossCursor));
		}
	}
}

#define TMP_NO_JOINT -1
#define TMP_PIVOT_JOINT -2
#define TMP_ROTATION_JOINT -3
int hlp_click_was_near_joint( QMouseEvent *e, GLView *g)
{
	ViewData &vd = *((ViewData *)(g->edata));
	vd.frm->mdl->rotate_bone = false;
	vd.frm->mdl->translate_bone = false;
	if( vd.frm->mdl->do_bone_bind) return TMP_NO_JOINT;
	if( vd.frm->mdl->sel_bone < 0) return TMP_NO_JOINT;
	
	const rgVector n( g->view_normal());
	rgVector v( e->x(), e->y(), 0);
	rggl_ortho_window_to_world( g->width(), g->height(),
		&v, 1, &(g->vmat.icamM));
	v = rgMakePlanar( v, n);

	rgVector *u = vd.frm->mdl->bone_build_v;
	const int bid = vd.frm->mdl->sel_bone;
	const int pj = vd.frm->mdl->mdl.skel.b[bid].pj;
	rgVector p( rgMakePlanar( u[pj], n));
	if( (v - p).length() < (vd.world_pixel_length * 16.0f))
		return TMP_PIVOT_JOINT;
	p = rgMakePlanar( u[(~pj) & 0x1], n);
	if( (v - p).length() < (vd.world_pixel_length * 16.0f))
		return TMP_ROTATION_JOINT;
	return TMP_NO_JOINT;
}

void vfn_mouse_down( QMouseEvent *e, GLView *g)
{
	ViewData &vd = *((ViewData *)(g->edata));
	sk3DRomModel &mdl = vd.frm->mdl->mdl;
	vd.mouse_down_x = e->x();
	vd.mouse_down_y = e->y();
	vd.mouse_moved = false;

	{ // Determine the length of one pixel in world space.
		rgVector v[2] = { rgVector( 0, 0, 0), rgVector( 1, 0, 0)};
		rggl_ortho_window_to_world( g->width(), g->height(),
				v, 2, &(g->vmat.icamM));
		vd.world_pixel_length = (v[0] - v[1]).length();
	}

	if( Qt::RightButton == e->button()) {
		// Cleanup current editing state or disable a current
		// editing action.
		switch( krn_get_edit_mode()) {
		case KED_MODE_SEL_VERT:
			if( vd.frm->mdl->sel_vertex >= 0) {
				vd.frm->mdl->sel_vertex = -1;
				g->setCursor( QCursor( Qt::arrowCursor));
				g->updateGL();
				return;
			}
			break;

		case KED_MODE_ADD_TRI:
			// Cancel triangle build if one is in progress.
			if( vd.frm->mdl->tri_build_count > 0) {
				vd.frm->mdl->tri_build_count = 0;
				g->updateGL();
				return;
			}
			break;
		}
		
		(vd.frm->m_menu[0])->popup( QCursor::pos());
		return;
	}

	switch( krn_get_edit_mode()) {
	case KED_MODE_SEL_BONE:
		switch( hlp_click_was_near_joint( e, g)) {
		case TMP_PIVOT_JOINT:
			vd.frm->mdl->translate_bone = true;
			vd.frm->mdl->sel_bone_rcs = mdl.skel.skout.b[
				vd.frm->mdl->sel_bone].rcs;
			break;

		case TMP_ROTATION_JOINT:
			vd.frm->mdl->rotate_bone = true;
			if( (Qt::ShiftButton & e->state()))
				vd.frm->mdl->twist_bone = true;
			else vd.frm->mdl->twist_bone = false;
			vd.frm->mdl->sel_bone_rcs = mdl.skel.skout.b[
				vd.frm->mdl->sel_bone].rcs;
			break;

		default:
			{
				rgVector v( e->x(), e->y(), 0);
				rggl_ortho_window_to_world( g->width(),
					g->height(), &v, 1, &(g->vmat.icamM));
				mdl.ortho_select_bone( v,
					g->view_normal(),
					mdl.skel.skout,
					vd.frm->mdl->sel_bone,
					vd.frm->mdl->sel_joint);
				rgVector *u = vd.frm->mdl->bone_build_v;
				mdl.get_absolute_bone_vertices(
					mdl.skel.skout,
					vd.frm->mdl->sel_bone,
					u[0], u[1]);
				vd.frm->mdl->update_all_views();
			}
			break;
		}
		break;

	case KED_MODE_ROTATE:
		krn_set_selection_center_point( vd);
		break;

	case KED_MODE_SCALE:
		krn_set_selection_center_point( vd);
		break;

	case KED_MODE_SEL_VERT:
		hlp_do_vert_sel_mouse_down( e, g, vd);
		break;

	case KED_MODE_SEL_BRUSH:
	case KED_MODE_ADD_BRUSH:
	case KED_MODE_ROTATE_BRUSH:
	case KED_MODE_SCALE_BRUSH:
		krn_do_brush_mouse_down( e, g, vd);
		break;

	case KED_MODE_ADD_VERT:
		{
			rgVector v( e->x(), e->y(), 0);
			rggl_ortho_window_to_world( g->width(), g->height(),
				&v, 1, &(g->vmat.icamM));
			vd.frm->mdl->add_vertex( vd.frm, v);
			vd.frm->mdl->modified = true;
			vd.frm->mdl->update_all_views();
		}
		break;

	case KED_MODE_CAMERA_TRAN:
		vd.do_cam_rotate = false;
		vd.do_cam_scale = false;
		if( (e->state() & Qt::ShiftButton)) vd.do_cam_scale = true;
		else if( (e->state() & Qt::ControlButton)) vd.do_cam_rotate = true;
		break;

	case KED_MODE_SEL_TRI:
		{
			rgVector v( e->x(), e->y(), 0);
			rggl_ortho_window_to_world( g->width(), g->height(),
				&v, 1, &(g->vmat.icamM));
			vd.frm->mdl->sel_triangle = mdl.ortho_select_tv_triangle(
				v, g->view_normal());
			vd.frm->mdl->set_selecting_view( vd.frm);
			vd.frm->mdl->update_all_views();
		}
		break;

	case KED_MODE_ADD_TRI:
		{
			rgVector v( e->x(), e->y(), 0);
			rggl_ortho_window_to_world( g->width(), g->height(),
				&v, 1, &(g->vmat.icamM));
			int vid = krn_select_vertex( *(vd.frm->mdl), v,
				g->view_normal(),
				vd.world_pixel_length * 16.0f);
			if( vid < 0) return;	// No point selected
			vd.frm->mdl->tri_build_mp = v;
			switch( vd.frm->mdl->tri_build_count) {
			case 0:	// No points added yet
				vd.frm->mdl->tri_build_vid[0] = vid;
				vd.frm->mdl->tri_build_count++;
				g->updateGL();
				break;

			case 1: // 1 point is currently part of the tri
				if( vid != vd.frm->mdl->tri_build_vid[0]) {
					vd.frm->mdl->tri_build_vid[1] = vid;
					vd.frm->mdl->tri_build_count++;
					g->updateGL();
				}
				break;

			case 2:	// 2 points are currently part of the tri
				// Build the triangle now.
				if( (vid != vd.frm->mdl->tri_build_vid[0]) &&
					(vid != vd.frm->mdl->tri_build_vid[1])) {

					vd.frm->mdl->mdl.add_triangle(
						vd.frm->mdl->tri_build_vid[0],
						vd.frm->mdl->tri_build_vid[1],
						vid);
					vd.frm->mdl->tri_build_vid[2] = vid;
					vd.frm->mdl->tri_build_count = 0;
					vd.frm->mdl->modified = true;
					vd.frm->mdl->update_all_views();
				}
				break;
			}
		}
		break;
	}
}
#undef TMP_NO_JOINT
#undef TMP_PIVOT_JOINT
#undef TMP_ROTATION_JOINT

float hlp_calc_bone_rot_angle( const ViewData &vd,
	const rgVector p_init, const rgVector p_fini,
	const rgVector n)
{
	rgVector p( vd.frm->mdl->bone_build_v[0]);
	sk3DRomModel &mdl = vd.frm->mdl->mdl;
	if( mdl.skel.b[ vd.frm->mdl->sel_bone].pj > 0)
		p = vd.frm->mdl->bone_build_v[1];
	p = rgMakePlanar( p, n);
	const rgVector a( rgMakePlanar( p_init, n) - p);
	const rgVector b( rgMakePlanar( p_fini, n) - p);
	const float ang = rgDotAngle( a, b);
	if( (a.cross( b)).dot( n) < 0.0f) return -ang;
	return ang;
}

void vfn_mouse_move( QMouseEvent *e, GLView *g)
{
	ViewData &vd = *((ViewData *)(g->edata));
	sk3DRomModel &mdl = vd.frm->mdl->mdl;
	vd.mouse_moved = true;
	vd.mouse_move_x = e->x();
	vd.mouse_move_y = e->y();

	// Movement when a mouse button is not pressed.
	// Movement on a right mouse button is not allowed.
	if( (e->state() & Qt::RightButton)) return;

	rgVector mwc[] = {	// Mouse World Coordinates
		rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
		rgVector( e->x(), e->y(), 0)};
	rggl_ortho_window_to_world( g->width(),
		g->height(), mwc, 2, &(g->vmat.icamM));
	const rgVector mdp( mwc[0]);	// Mouse Down Point (World coord)
	const rgVector mcp( mwc[1]);	// Mouse Current Point (World coord)

	krn_print_location( krn_snap_to_grid( vd, mcp));

	if( !(e->state() & Qt::LeftButton)) {
		switch( krn_get_edit_mode()) {
		case KED_MODE_SEL_VERT:
			if( vd.frm->mdl->sel_vertex >= 0) {
				const int vid = vd.frm->mdl->sel_vertex;
				const rgVector u( rgMakePlanar(
					mdl.get_tv_vertex( vid), g->view_normal()));
				if( (mcp - u).length() <
					(vd.world_pixel_length * 16.0f)) {
					g->setCursor( QCursor( Qt::CrossCursor));
				} else g->setCursor( QCursor( Qt::ArrowCursor));
			} else g->setCursor( QCursor( Qt::ArrowCursor));
			break;

		case KED_MODE_ADD_TRI:
			if( vd.frm->mdl->tri_build_count > 0) {
				vd.frm->mdl->tri_build_mp = mcp;
				g->updateGL();
			}
			break;
		}
		return;
	}

	// Movement when a mouse button is pressed.
	switch( krn_get_edit_mode()) {
	case KED_MODE_ROTATE:
		{
			int ang_lock = -1;
			if( (Qt::ShiftButton & e->state())) ang_lock = 15;
			vd.frm->mdl->tool_matrix = krn_gen_rotation_matrix(
				mdp, mcp, vd.frm->mdl->selection_center_point,
				g->view_normal(), ang_lock);
		}
		vd.frm->mdl->update_all_views();
		break;

	case KED_MODE_SCALE:
		vd.frm->mdl->tool_matrix = krn_gen_scale_matrix( g->height(),
			rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
			rgVector( e->x(), e->y(), 0),
			vd.frm->mdl->selection_center_point,
			2.0f);
		vd.frm->mdl->update_all_views();
		break;

	case KED_MODE_SEL_BRUSH:
	case KED_MODE_ADD_BRUSH:
	case KED_MODE_ROTATE_BRUSH:
	case KED_MODE_SCALE_BRUSH:
		krn_do_brush_mouse_move( e, g, vd);
		break;

	case KED_MODE_SEL_BONE:
		if( vd.frm->mdl->rotate_bone) {
			mdl.skel.skout.b[ vd.frm->mdl->sel_bone].rcs =
				vd.frm->mdl->sel_bone_rcs;
			rgVector n( g->view_normal());
			if( vd.frm->mdl->twist_bone) {
				n = (mdl.skel.get_full_bone_rcs(
					vd.frm->mdl->sel_bone,
					mdl.skel.skout)).up;
			}
			mdl.skel.rotate_rcs( vd.frm->mdl->sel_bone,
				mdl.skel.skout,
				hlp_calc_bone_rot_angle( vd, mdp, mcp,
					g->view_normal()),
				n);
			mdl.transform_skel_frame();
			mdl.get_absolute_bone_vertices(
				mdl.skel.skout,
				vd.frm->mdl->sel_bone,
				vd.frm->mdl->bone_build_v[0],
				vd.frm->mdl->bone_build_v[1]);
			vd.frm->mdl->update_all_views();

		} else if( vd.frm->mdl->translate_bone) {
			const rgVector v[] = {
				krn_snap_to_grid( vd, mdp),
				krn_snap_to_grid( vd, mcp)};
			mdl.skel.skout.b[ vd.frm->mdl->sel_bone].rcs =
				vd.frm->mdl->sel_bone_rcs;
			mdl.skel.translate_rcs( vd.frm->mdl->sel_bone,
				mdl.skel.skout,
				v[1] - v[0]);
			mdl.transform_skel_frame();
			mdl.get_absolute_bone_vertices(
				mdl.skel.skout,
				vd.frm->mdl->sel_bone,
				vd.frm->mdl->bone_build_v[0],
				vd.frm->mdl->bone_build_v[1]);
			vd.frm->mdl->update_all_views();
		}
		break;

	case KED_MODE_ADD_BONE:
		// Use the selection boxes vertex to store the second bone
		// vertex.
		{
			rgVector *v = vd.frm->mdl->bone_build_v;
			v[0] = krn_snap_to_grid( vd, mdp);
			v[1] = krn_snap_to_grid( vd, mcp);
			vd.frm->mdl->show_bone_cursor = true;
			g->updateGL();
		}
		break;

	case KED_MODE_SEL_VERT:
		if( vd.do_vertex_move) {
			vd.show_vertex_sel_box = false;
			if( (vd.frm->mdl->sel_vertex >= 0)) {
				const int vid = vd.frm->mdl->sel_vertex;
				const rgVector &n = g->view_normal();
				rgVector v( mcp + (n.dot( mdl.get_tv_vertex( vid)) * n));
				mdl.v[vid] = krn_snap_to_grid( vd, v);
				mdl.set_tv_vertex( vid, mdl.v[vid]);
				vd.frm->mdl->modified = true;
				g->updateGL();
			}

		} else if( vd.do_sel_box_move) {
			rgVector v( e->x(), e->y(), 0);
			rggl_ortho_window_to_world( g->width(), g->height(),
				&v, 1, &(g->vmat.icamM));
			Model &m = *(vd.frm->mdl);
			v = krn_snap_to_grid( vd, v);
			m.sel_box.set_offset( v - m.sel_box.mouse_down());
			g->updateGL();

		} else {
		// Show the sizing of the vertex selection box. Also, disable
		// point selection.
			vd.frm->mdl->sel_vertex = -1;
			vd.frm->mdl->sel_box.set_box(
				vd.mouse_down_x, vd.mouse_down_y,
				e->x(), e->y(),
				g);
			vd.show_vertex_sel_box = true;
			g->updateGL();
		}
		break;

	case KED_MODE_ADD_TRI:
		{
			rgVector v( e->x(), e->y(), 0);
			rggl_ortho_window_to_world( g->width(), g->height(),
				&v, 1, &(g->vmat.icamM));
			vd.frm->mdl->tri_build_mp = v;
			g->updateGL();
		}
		break;

	case KED_MODE_CAMERA_TRAN:
		if( vd.do_cam_rotate) {
			g->vmat.do_view_3way_rotate( g->width(), g->height(),
				rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0), 0, 0, false);
			g->updateGL();

		} else if( vd.do_cam_scale) {
			g->vmat.do_view_scale( g->height(),
				rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0), 2.0f, 0, 0, false);
			g->updateGL();	

		} else {
			g->vmat.do_view_translate( g->width(), g->height(),
				rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0), 0, 0, false);
			g->updateGL();
		}
		break;
	}
}

void vfn_mouse_up( QMouseEvent *e, GLView *g)
{
	ViewData &vd = *((ViewData *)(g->edata));

	switch( krn_get_edit_mode()) {
	case KED_MODE_ROTATE:
	case KED_MODE_SCALE:
		krn_set_transformed_selection_box( vd);
		vd.frm->mdl->update_all_views();
		break;

	case KED_MODE_ADD_BONE:
		if( vd.frm->mdl->show_bone_cursor) {
			// Enforce a minimum length for the bones.
			rgVector *v = vd.frm->mdl->bone_build_v;
			if( ((v[0] - v[1]).length()) > KRN_MIN_BONE_LENGTH) {
				vd.frm->mdl->mdl.add_bone( v[0], v[1]);
			}
			vd.frm->mdl->show_bone_cursor = false;
			vd.frm->mdl->update_all_views();
		}
		break;

	case KED_MODE_SEL_VERT:

		// Do point/vertex selection only if the user did a fixed
		// click. If they moved the mouse, then the user was doing
		// an area selection, not a single point/vertex seleciton.
		if( !(vd.do_vertex_move)) {
			if( !(vd.mouse_moved)) {
			// Do a single point/vertex selection.
				vd.frm->mdl->clear_verticies_selection();
				rgVector v( e->x(), e->y(), 0);
				rggl_ortho_window_to_world( g->width(), g->height(),
					&v, 1, &(g->vmat.icamM));
	
				vd.frm->mdl->sel_vertex = krn_select_vertex(
					*(vd.frm->mdl), v, g->view_normal(),
					vd.world_pixel_length * 16.0f);
				if( vd.frm->mdl->sel_vertex >= 0)
					g->setCursor( QCursor( Qt::CrossCursor));
				vd.frm->mdl->update_all_views();
			
			} else if( vd.do_sel_box_move &&
				(vd.frm->mdl->sel_verticies_count > 0)) {
				// Move all selected vertices.
				Model &m = *(vd.frm->mdl);
				vd.frm->mdl->mdl.offset_tv_verticies(
					m.sel_verticies, m.sel_verticies_count,
					m.sel_box.offset());
				vd.frm->mdl->mdl.set_vertex_from_tv(
					m.sel_verticies, m.sel_verticies_count);

				rgMatrix t;
				t.translate( vd.frm->mdl->sel_box.offset());
				vd.frm->mdl->sel_box.transform( t);
				vd.frm->mdl->sel_box.set_offset( rgVector());
				vd.do_sel_box_move = false;

				vd.frm->mdl->modified = true;
				vd.frm->mdl->update_all_views();

			} else {
			// Do an area selection.
				krn_select_vertices( *(vd.frm->mdl),
					g->view_normal());
				if( !(vd.frm->mdl->sel_vertices))
					vd.show_vertex_sel_box = false;
				vd.frm->mdl->update_all_views();
			}

		} else {
			// A single vertex is being moved.
			if( (vd.frm->mdl->sel_vertex >= 0)) {
				vd.frm->mdl->mdl.set_vertex_from_tv(
					&(vd.frm->mdl->sel_vertex), 1);
			}
		}
		break;

	case KED_MODE_CAMERA_TRAN:
		if( vd.do_cam_rotate) {
			g->vmat.do_view_3way_rotate( g->width(), g->height(),
				rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0), 0, 0, true);
			g->updateGL();

		}  else if( vd.do_cam_scale) { // Perform the scale now.
			g->vmat.do_view_scale( g->height(),
				rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0),
				2.0f, 0, 0, true);
			g->updateGL();

		} else { // Perform the translation now.
			g->vmat.do_view_translate( g->width(), g->height(),
				rgVector( vd.mouse_down_x, vd.mouse_down_y, 0),
				rgVector( e->x(), e->y(), 0), 0, 0, true);
			g->updateGL();
		}
		break;
	}
}

void hlp_hide_selected_verticies( ViewData &vd)
{
	int *v = vd.frm->mdl->sel_verticies;
	for( int i = 0; i < (vd.frm->mdl->sel_verticies_count); i++)
		vd.frm->mdl->mdl.v[ v[i]].hide();
}

void vfn_key_down( QKeyEvent *e, GLView *g)
{
	ViewData &vd = *((ViewData *)(g->edata));
	Model &kmdl = *(vd.frm->mdl);

	switch( krn_get_edit_mode()) {
	case KED_MODE_SEL_BRUSH:
	case KED_MODE_ADD_BRUSH:
		switch( e->key()) {
		case Qt::Key_Delete:
			kmdl.mdl.m_csg->delete_brush( kmdl.mdl.m_brush_cursor);
			kmdl.mdl.m_brush_cursor = 0;
			kmdl.update_all_views();
			break;
		}
		break;

	case KED_MODE_SCALE_BRUSH:
		switch( e->key()) {
		case Qt::Key_X: kmdl.scale_brush_mode = KRN_SCALE_BRUSH_X; break;
		case Qt::Key_Y: kmdl.scale_brush_mode = KRN_SCALE_BRUSH_Y; break;
		case Qt::Key_Z: kmdl.scale_brush_mode = KRN_SCALE_BRUSH_Z; break;
		default: kmdl.scale_brush_mode = KRN_SCALE_BRUSH_XYZ; break;
		}
		break;

	case KED_MODE_SEL_BONE:
		switch( e->key()) {
		case Qt::Key_H:
			kmdl.mdl.skel.reset_bone(
				vd.frm->mdl->sel_bone,
				vd.frm->mdl->mdl.skel.skout);
			kmdl.mdl.transform_skel_frame();
			kmdl.mdl.get_absolute_bone_vertices(
				kmdl.mdl.skel.skout,
				kmdl.sel_bone,
				kmdl.bone_build_v[0],
				kmdl.bone_build_v[1]);
			kmdl.update_all_views();
			break;

		case Qt::Key_Home:
			vd.frm->mdl->mdl.skel.reset_bone_and_slaves(
				vd.frm->mdl->sel_bone,
				vd.frm->mdl->mdl.skel.skout);
			vd.frm->mdl->mdl.transform_skel_frame();
			vd.frm->mdl->mdl.get_absolute_bone_vertices(
				vd.frm->mdl->mdl.skel.skout,
				vd.frm->mdl->sel_bone,
				vd.frm->mdl->bone_build_v[0],
				vd.frm->mdl->bone_build_v[1]);
			vd.frm->mdl->update_all_views();
			break;

		case Qt::Key_Escape:
			if( vd.frm->mdl->do_bone_bind) {
				vd.frm->mdl->do_bone_bind = false;
				vd.frm->mdl->update_all_views();
			}
			break;

		case Qt::Key_Delete:
			if( vd.frm->mdl->sel_bone >= 0) {
				vd.frm->mdl->do_bone_bind = false;
				vd.frm->mdl->mdl.delete_bone(
					vd.frm->mdl->sel_bone,
					&(vd.frm->mdl->mdl.skel.skout));
				vd.frm->mdl->sel_bone = -1;
				vd.frm->mdl->update_all_views();
			}
			break;

		case Qt::Key_M:
			if(	vd.frm->mdl->do_bone_mount &&
				(vd.frm->mdl->sel_bone >= 0)) {
				vd.frm->mdl->mdl.skel.mount_bone(
					vd.frm->mdl->sel_bone,
					vd.frm->mdl->bone_bind_slave,
					vd.frm->mdl->mdl.skel.skout);
				vd.frm->mdl->do_bone_mount = false;
				vd.frm->mdl->update_all_views();

			} else {
				if( vd.frm->mdl->sel_bone >= 0) {
					vd.frm->mdl->bone_build_v[2] =
						vd.frm->mdl->bone_build_v[0];
					vd.frm->mdl->bone_build_v[3] =
						vd.frm->mdl->bone_build_v[1];
					vd.frm->mdl->bone_bind_slave =
						vd.frm->mdl->sel_bone;
					vd.frm->mdl->do_bone_mount = true;
					vd.frm->mdl->update_all_views();
				}
			}
			break;

		case Qt::Key_B:
			if(	vd.frm->mdl->do_bone_bind &&
				(vd.frm->mdl->sel_bone >= 0)) {
				vd.frm->mdl->mdl.skel.bind_bone_to_bone(
					vd.frm->mdl->sel_bone,
					vd.frm->mdl->bone_bind_slave);
				vd.frm->mdl->do_bone_bind = false;
				vd.frm->mdl->update_all_views();

			} else {
				if( vd.frm->mdl->sel_bone >= 0) {
					vd.frm->mdl->bone_build_v[2] =
						vd.frm->mdl->bone_build_v[0];
					vd.frm->mdl->bone_build_v[3] =
						vd.frm->mdl->bone_build_v[1];
					vd.frm->mdl->bone_bind_slave =
						vd.frm->mdl->sel_bone;
					vd.frm->mdl->do_bone_bind = true;
					vd.frm->mdl->update_all_views();
				}
			}
			break;
		}
		break;

	case KED_MODE_SEL_TRI:
		switch( e->key()) {
		case Qt::Key_F:
			if( vd.frm->mdl->sel_triangle >= 0) {
				vd.frm->mdl->mdl.f[
					vd.frm->mdl->sel_triangle].flip();
				vd.frm->mdl->update_all_views();
			}
			break;

		case Qt::Key_I:
			if( (Qt::ShiftButton & e->state())) {
				bool ok;
				const int id = QInputDialog::getInteger(
					"Highlight Face", "Enter face id:",
						vd.frm->mdl->sel_triangle,
						0,
						vd.frm->mdl->mdl.face_count() -1,
						1,
						&ok,
						vd.frm);
				if( ok) {
					vd.frm->mdl->sel_triangle = id;
					vd.frm->mdl->update_all_views();
				} else return;

			} else if( vd.frm->mdl->sel_triangle >= 0) {
				std::cout << "Face id = "
					<< vd.frm->mdl->sel_triangle
					<< ", "
					<< (vd.frm->mdl->mdl.f[
						vd.frm->mdl->sel_triangle])
					<< "\n";
			}
			break;

		case Qt::Key_Delete:
			if( vd.frm->mdl->sel_triangle >= 0) {
				vd.frm->mdl->mdl.delete_face(
					vd.frm->mdl->sel_triangle);
				vd.frm->mdl->sel_triangle = -1;
				vd.frm->mdl->update_all_views();
			}
			break;
		}
		break;

	case KED_MODE_SEL_VERT:
		switch( e->key()) {
		// Delete selected vertex or group of selected vertices.
		case Qt::Key_Delete:
			if( !(vd.do_sel_box_move)) {
				if( vd.show_vertex_sel_box) {
					std::cerr << "Delete multi\n";
				} else if( vd.frm->mdl->sel_vertex >= 0) {
					vd.frm->mdl->mdl.delete_vertex(
						vd.frm->mdl->sel_vertex);
					vd.frm->mdl->sel_vertex = -1;
					vd.frm->mdl->mdl.transform_skel_frame();
					vd.frm->mdl->modified = true;
				}
			}
			vd.frm->mdl->update_all_views();
			break;

		// Hide all selected vertices.
		case Qt::Key_H:
			if( vd.frm->mdl->sel_verticies) {
				hlp_hide_selected_verticies( vd);
				// Hidden vertices cannot be selected.
				vd.frm->mdl->clear_verticies_selection();
				vd.show_vertex_sel_box = false;
				vd.frm->mdl->modified = true;
				vd.frm->mdl->update_all_views();
			}
			break;

		// Reveal all vertices.
		case Qt::Key_R:
			// FIXME: Model is marked as modified even if there
			// are no hidden vertices to reveal.
			vd.frm->mdl->mdl.reveal_all_vertices();
			vd.frm->mdl->modified = true;
			vd.frm->mdl->update_all_views();
			break;

		case Qt::Key_W:
			if( vd.frm->mdl->sel_vertex < 0) break;
			std::cerr << "v[" << vd.frm->mdl->sel_vertex << "].weight = "
				<< vd.frm->mdl->mdl.v[ vd.frm->mdl->sel_vertex].
					weight << "\n";
			break;
		}
		break;
	}
}

void hlp_draw_build_triangle( Model &mdl)
{
	if( 0 == mdl.tri_build_count) return;
	glLineWidth( 5);
	glColor3f( 0, 0, 1);
	glBegin( GL_LINE_STRIP);
	switch( mdl.tri_build_count) {
	case 2:
		glVertex( mdl.mdl.get_tv_vertex( mdl.tri_build_vid[0]));
		glVertex( mdl.mdl.get_tv_vertex( mdl.tri_build_vid[1]));
		glVertex( mdl.tri_build_mp);
	case 1:
		glVertex( mdl.mdl.get_tv_vertex( mdl.tri_build_vid[0]));
		glVertex( mdl.tri_build_mp);
		break;
	}
	glEnd();
	glLineWidth( 1);
}

void hlp_draw_build_bone( const Model &mdl)
{
	if( !(mdl.show_bone_cursor)) return;
	const rgVector *v = mdl.bone_build_v;
	if( ((v[0] - v[1]).length()) > KRN_MIN_BONE_LENGTH) {
		glLineWidth( 5);	// Draw bone
		glBegin( GL_LINES);
		glColor3f( 0, 0, 1);
		glVertex( v[0]);
		glVertex( v[1]);
		glEnd();
		glLineWidth( 1);
	}

	glPointSize( 5);	// Draw bone joints
	glBegin( GL_POINTS);
	glColor3f( 0, 0.75, 0);
	glVertex( v[0]);
	glVertex( v[1]);
	glEnd();
}
