/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#ifndef RG3DROM_CSG_H
#define RG3DROM_CSG_H

#include <rgCSG.h>

// XXX: "rg3DRom" brush identifiers (identifiers 1001 through 2000 inclusive,
//	are reserved for "rg3DRom").
#define RG3DROM_CSG_CUBE	1001

class sk3DRomModel;

class sk3DRomCSGCube : public rgCSGCube
{
public:
	sk3DRomCSGCube() {
		m_model = 0;
		m_flags = 0;
		m_bone_id = -1;
	}
	sk3DRomCSGCube( sk3DRomModel *p_mdl) {
		m_model = p_mdl;
		m_flags = 0;
		m_bone_id = -1;
	}

	sk3DRomModel *m_model;

	int m_flags;
	int m_bone_id;

	int type() const { return RG3DROM_CSG_CUBE;}
	int version() const { return 0;}
	int save_data( std::ostream &des) const;
	int load_data( std::istream &src);

	void draw();
	void real_draw_cursor( const rgVector vc, const rgVector ec);
	rgVector position() const {
		if( !bone_id_valid()) return m_pos.p;
		return (m_model->skel.b[ m_bone_id].fcs).map_point( m_pos.p);
	}
	rgCoordSystem coord_system() const {
		if( !bone_id_valid()) return m_pos;
		return (m_model->skel.b[ m_bone_id].fcs).map_csys( m_pos);
	}

	bool selectable() const {
		if(	group_id_valid() &&
			(m_model->group[ m_group_id]).hidden()) return false;
		return true;
	}

	void gen_verts( std::vector <rgVector> &v) const;

protected:

private:
	sk3DRomCSGCube( const sk3DRomCSGCube &);
	sk3DRomCSGCube &operator =( const sk3DRomCSGCube &);

	int load_version_0_data( std::istream &src);
	bool bone_id_valid() const {
		if( !m_model) return false;
		if( m_bone_id < 0) return false;
		if( m_bone_id >= int( m_model->skel.b.size())) return false;
		return true;
	}
	void dbg_bone_loss_check();

	bool group_id_valid() const {
		if( !m_model) return false;
		if( m_group_id < 0) return false;
		if( m_group_id >= int( m_model->group.size())) return false;
		return true;
	}
};

#endif
