/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <rgPort.h>
#include <iostream>
#include <string.h>
#include <math.h>
#include "rg3DRom.h"

void interpolate_bone_csys( const rgFloat w, const rgCoordSystem &src_cs,
	const rgCoordSystem &des_cs, rgCoordSystem &res_cs)
{
	res_cs.p = (src_cs.p * (1.0f - w)) + (des_cs.p * w);
	const rgVector hup( ((src_cs.up + des_cs.up) * 0.5f).normalize());
	const rgVector hside( ((src_cs.side + des_cs.side) * 0.5f)
		.normalize());
	if( w < 0.5f) {
		const float d = w * 2.0f;
		const float s = 1.0f - d;
		res_cs.up =
			((src_cs.up * s) + (hup * d)).normalize();
		res_cs.side =
			((src_cs.side * s) + ( hside * d)).normalize();

	} else {
		const float d = (w - 0.5f) * 2.0f;
		const float s = 1.0f - d;
		res_cs.up =
			((hup * s) + (des_cs.up * d)).normalize();
		res_cs.side =
			((hside * s) + (des_cs.side * d)).normalize();
	}
	res_cs.eye = res_cs.side.cross( res_cs.up);
}


void sk3DRomSkel::inbetween_skeleton( const float w,
	const sk3DRomSkelFrame &ska, const sk3DRomSkelFrame &skb)
{
	inbetween_skeleton( w, ska, skb, skout);
}

void sk3DRomSkel::inbetween_skeleton( const float w,
	const sk3DRomSkelFrame &ska, const sk3DRomSkelFrame &skb,
	sk3DRomSkelFrame &skc)
{
	// IMPORTANT NOTE: The skeleton interpolation can be done
	//	bone-by-bone in any arbitrary order. This is because every
	//	bone is controlled by a coordinate system. By simply
	//	interpolating the coordinate systems, the entire skeleton
	//	itself will be interpolated. Transformation and rendering
	//	functions gaurentee that bones will remain linked properly
	//	throughout the interpolation (so there is no need to check
	//	for that in this function).
	//
	//	Also, the ability for bone-by-bone transformation means that
	//	the output skeleton "skc" can be the same as "ska" or "skb"
	//	(that is "skc" has the same address as "ska" or "skb").

	const int istop = int( ska.b.size());
	for( int i = 0; i < istop; i++) {
		interpolate_bone_csys( w, ska.b[i].rcs, skb.b[i].rcs,
			skc.b[i].rcs);
	}
}

void sk3DRomSkel::copy_mount_data( sk3DRomSkelFrame &des,
	const sk3DRomSkelFrame &src) const
{
	if( src.b.size() != des.b.size()) return;
	for( int i = 0; i < int( src.b.size()); i++) {
		des.b[i].msb = src.b[i].msb;
		des.b[i].mmb = src.b[i].mmb;
		des.b[i].cms = src.b[i].cms;
	}
}
