/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RGNETWORK_H
#define RGNETWORK_H

//#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
//#include <errno.h>
//#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
//#include <netinet/in.h>
#include <arpa/inet.h>

#define RG_SOCK_STREAM SOCK_STREAM
#define RG_SOCK_DGRAM SOCK_DGRAM
#define RG_SOCK_SEQPACKET SOCK_SEQPACKET
#define RG_SOCK_RAW SOCK_RAW
#define RG_SOCK_RDM SOCK_RDM

// FIXME: Helper functions have been provided because the compiler cannot tell
//	tell the difference between "rgSocket::send" and "send" (as well as
//	other functions) from the C library (even if parameter list is
//	non-ambiguous).
inline int rgsocket_send( const int s, const void *msg, const size_t len,
	const int flags)
{ return send( s, msg, len, flags);}

inline int rgsocket_close( int fd)
{ return close( fd);}

inline int rgsocket_accept( int s, struct sockaddr *addr, socklen_t *addrlen)
{ return accept( s, addr, addrlen);}

inline int rgsocket_connect( int s, const struct sockaddr *serv_addr,
	socklen_t addr_len)
{ return connect( s, serv_addr, addr_len);}

class rgSocket
{
public:

	rgSocket();
	~rgSocket();

	//rgSocket &operator =( const rgSocket &s) {
	//	close();
	//	do_copy( s);
	//	return *this;
	//}

	const int port() const { return m_port;}

	bool is_good() const { return m_good;}
	bool connected() const {
		if( m_remotefd > 0) return true;
		return false;
	}

	int open_listen_port( const int p_port) {
		return open_listen_port( p_port, RG_SOCK_STREAM);
	}
	int open_listen_port( const int p_port, const int p_type);

	int connect( const char *ip, const int p_port) {
		return connect( ip, p_port, RG_SOCK_STREAM);
	}
	int connect( const char *ip, const int p_port, const int p_type);

	int accept();

	void close() {
		close_local();
		close_remote();
		m_good = false;
	}

	void close_local() {
		if( m_sockfd) { rgsocket_close( m_sockfd); m_sockfd = 0;}
	}
	void close_remote() {
		if( m_remotefd) { rgsocket_close( m_remotefd); m_remotefd = 0;}
	}

	const char *remote_ip() const {
		return inet_ntoa( m_remote_addr.sin_addr);
	}

	int send( const void *msg, const int len) const {
		return rgSocket::send( msg, len, 0);
	}
	int send( const void *msg, const int len, const int flags) const {
		if( !is_good()) return -1;
		return rgsocket_send( m_remotefd, msg, len, flags);
	}

	int receive( void *buf, const int len) const {
		return receive( buf, len, 0);
	}
	int receive( void *buf, const int len, const int flags) const {
		if( !is_good()) return -1;
		return recv( m_remotefd, buf, len, flags);
	}

	int local_fd() const { return m_sockfd;}
	int remote_fd() const { return m_remotefd;}

private:

	bool m_good;
	int m_sockfd;
	int m_port;
	int m_type;
	int m_optval;
	int m_backlog;
	int m_remotefd;
	sockaddr_in m_local_addr;
	sockaddr_in m_remote_addr;

	/* XXX: Copying of rgSockets is not allowed.
	void do_copy( const rgSocket &s) {
		m_backlog = s.m_backlog;
		m_type = s.m_type;
                m_port = s.m_port;
                m_local_addr = s.m_local_addr;
                m_remote_addr = s.m_remote_addr;
	}
	*/

	// XXX: Copying of rgSockets is not allowed.
	rgSocket( const rgSocket &s);
	rgSocket &operator =( const rgSocket &s);
};

#endif
