/*
Copyright (C) 2000-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <iostream>
#include "rgOpenGL.h"
#include "rgError.h"
#include <GL/glu.h>
#include <vector>


//#define RGGLTEXTURE_DEBUG

std::vector <rgGLTexture_ref_tag> rgGLTexture::generated_textures;

bool rgGLTexture::find_texture_ref( const GLuint id, int *p_i)
{
	int &i = *p_i;
	const int istop = int( generated_textures.size());
	for( i = 0; i < istop; i++)
		if( id == generated_textures[i].id) return true;

	return false;
}

void rgGLTexture::add_texture_reference( const GLuint id)
{
	if( !id) return;
	#ifdef RGGLTEXTURE_DEBUG
	std::cerr << "Adding texture reference " << id << "\n";
	#endif
	generated_textures.push_back( rgGLTexture_ref_tag( id, 1));
}

void rgGLTexture::inc_texture_reference( const GLuint id)
{
	// No need to increment a texture reference if the texture ID is
	// zero.
	if( !id) return;

	int i;
	if( find_texture_ref( id, &i)) {
		generated_textures[i].count++;
		return;
	}
	#ifdef RGGLTEXTURE_DEBUG
	std::cerr << "OpenGL texture generated but not found in list!!!\n"
		"ID = " << id << "\n";
	#endif
}

void rgGLTexture::dec_texture_reference( const GLuint id)
{
	int i;
	if( find_texture_ref( id, &i)) { 
		generated_textures[i].count--;
		if( !(generated_textures[i].count)) {
			glDeleteTextures( 1, &id);
			generated_textures.erase(
				generated_textures.begin() + i);
			#ifdef RGGLTEXTURE_DEBUG
			std::cerr << "Deleting texture " << id << "!!!\n";
			#endif
		}
		return;
	}
	#ifdef RGGLTEXTURE_DEBUG
	std::cerr << "OpenGL texture to delete not found in list!!!\n"
		"ID = " << id << "\n";
	#endif
}

rgGLTexture::rgGLTexture()
{
	colorKey = RG_RGB( 0, 0, 0);
	textureID = 0;

	width = height = 0;
}



rgGLTexture::rgGLTexture( const rgGLTexture &t)
{
	#ifdef RGGLTEXTURE_DEBUG
	std::cerr << "rgGLTexture::rgGLTexture( const rgGLTexture &t)\n";
	#endif
	inc_texture_reference( t.textureID);

	copy_data( t);
/*
	if( !texture_semaphore) {
		texture_semaphore = new int;
		// NOTE: "this" and "t" are referencing the texture, therefore
		//	there are two references.
		*texture_semaphore = 2;
	} else {
		// Add texture "t" as a new reference.
		(*texture_semaphore)++;
	}
*/
}

rgGLTexture &rgGLTexture::operator =( const rgGLTexture &t)
{
	#ifdef RGGLTEXTURE_DEBUG
	std::cerr << "rgGLTexture &rgGLTexture::operator =( rgGLTexture &t)\n";
	#endif
	// IMPORTANT: Before aquiring the new texture "t", delete the current
	//	texture data.
	clear();
	inc_texture_reference( t.textureID);
	copy_data( t);
/*
	if( !(t.texture_semaphore)) {
		t.texture_semaphore = new int;
		(*texture_semaphore) = 1;
	}

	textureID = t.textureID;
	width = t.width;
	height = t.height;
	colorKey = t.colorKey;
	texture_semaphore = t.texture_semaphore;
	// Add "this" texture as a new reference.
	(*texture_semaphore)++;
*/

	return *this;
}

void rgGLTexture::copy_data( const rgGLTexture &t)
{
	width = t.width;
	height = t.height;

	colorKey = t.colorKey;
	textureID = t.textureID;
}


rgGLTexture::~rgGLTexture()
{
	clear();
}

/*
void rgGLTexture::add_reference( rgGLTexture &t)
{
	if( !texture_semaphore) {
                texture_semaphore = new int;
                // NOTE: "this" and "t" are referencing the texture, therefore
                //      there are two references.
                *texture_semaphore = 2;
		
        } else {
                // Add texture "t" as a new reference.
                (*texture_semaphore)++;
        }
}
*/

void rgGLTexture::clear()
{
	/* Delete the current loaded texture (if any) */
	if( textureID) {
		dec_texture_reference( textureID);
		/*
		if( texture_semaphore) {
			int &t = (*texture_semaphore);
			if( t > 1) {
				t--;
				std::cerr << t
					<< "references remain, keep tex!!!\n";
				return;
			} else {
				delete texture_semaphore;
				texture_semaphore = 0;
				return;
			}
		}
		*/
/*
		// Break links with other textures.
		if( prev_ref) prev_ref->next_ref = next_ref;
		if( next_ref) next_ref->prev_ref = prev_ref;

		// If there were no links then delete the texture (since
		// it will no longer be used/referenced).
		if( (!prev_ref) && (!next_ref)) {
			std::cerr << (void *)this <<
				" - Deleting texture!!!!\n";



			glDeleteTextures( 1, &textureID);
		} else {
			std::cerr << "Texture " << textureID << " is still "
				" being referenced.\n";
		}
*/
		textureID = 0;
	}

	width = 0;
	height = 0;
}

int rgGLTexture::load( const char *fname)
{
	return load( fname, 0);
}

void rgGLTexture::setColorKey( int r, int g, int b)
{
	colorKey = RG_RGB( r, g, b);
}

int rgGLTexture::load( const char *fname, int flags)
{
	if( !fname) return RGERR_BAD_PARAM;

	/********************************************************************
	 * Check the given flags.
	 ********************************************************************/
	/* RGGL_RGB_ALPHA_EXPAND takes precedence over RGGL_COLOR_KEY. */
	if( (flags & RGGL_RGB_ALPHA_EXPAND)) flags &= ~RGGL_COLOR_KEY;
	
	/* Delete the current loaded texture (if any) */
	clear();

	/********************************************************************
	 * Loads a texture represented as a BMP file.
	 *
	 * There is no restriction placed on the dimensions of the image
	 * except that the dimensions must be positive and not greater then
	 * RGGL_MAX_TEX_WIDTH and RGGL_MAX_TEX_HEIGHT. Although OpenGL requries
	 * textures that have dimensions that are a power of 2, OpenGL will
	 * itself correct textures that do not meet this requirment. So, this
	 * function does not enforce that textures should have dimensions
	 * that are a power of 2.
	 ********************************************************************/
	rgBMP bmp;
	int bmpFlags = RGBMP_FORCE_RGB;	/* OpenGL requires at least RGB for
					 * good performance.
					 */

	/* If the caller requested a texture with an alpha component, then
	 * a bitmap in RGBA format is required.
	 */
	if( (flags & (RGGL_RGB_ALPHA_EXPAND | RGGL_COLOR_KEY)))
		bmpFlags |= RGBMP_FORCE_RGBA;
	
	int error;
	if( (flags & RGGL_LARGE_TEXTURE)) {
		// Do not restrict the size of textures.
		error = bmp.load( fname, bmpFlags);
	} else {
		// Restrict the size of textures.
		error = bmp.load( fname, bmpFlags, RGGL_MAX_TEX_WIDTH,
			RGGL_MAX_TEX_HEIGHT);
	}
	if( error != RGERR_SUCCESS) return error;


	/* Bitmaps come in BGR order however RGB order is required. */
	bmp.convertBGRtoRGB();
	width = bmp.getWidth();
	height = bmp.getHeight();

	/* rgBMP will load the picture right-side-up, however OpenGL expects
	 * the pictures to be upside down.
	 */
	bmp.flip_vertically();

	/* Get a new texture object from opengl. */
	glGenTextures( 1, &textureID);
	/* If a valid name was not generated then an error occured. The error
	 * might be that there is no current rendering context.
	 */
	if( !textureID) return -1;

	glBindTexture( GL_TEXTURE_2D, textureID);

	/* Select how this texture is to be mapped on to primitives. */
	/* TODO: Support for mip-maps should be added. */
	if( (flags & RGGL_MIPMAP)) {
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			GL_LINEAR);
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_LINEAR_MIPMAP_NEAREST);

	} else if( (flags & RGGL_SMOOTH_TEXTURE)) {
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_LINEAR);
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			GL_LINEAR);
	} else {
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_NEAREST);
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			GL_NEAREST);
	}

	/* The created texture will support "wrapping" of its coordinates.
	 * This allows an image to be textured repeatedly across a primitive.
	 */
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	/* Load the image data into OpenGL. */
	GLint inpFormat = 3;	/* RGB */
	GLint outFormat = GL_RGB;
	if( (flags & (RGGL_RGB_ALPHA_EXPAND | RGGL_COLOR_KEY))) { 
		inpFormat = 4;	/* RGBA */
		outFormat = GL_RGBA;
	}

	/* Perform any requested alpha calculations now. */
	if( (flags & RGGL_RGB_ALPHA_EXPAND)) {
		/* Automatically create the alpha channel components from the
		 * image's RGB values (done for every pixel).
		 */
		genAlphaFromRGB( bmp);
		
	} else if( (flags & RGGL_COLOR_KEY)) {
		genAlphaFromColorKey( bmp);
	}

	if( (flags & RGGL_MIPMAP)) {
		GLint internalformat = GL_RGB;
		if( (flags & (RGGL_RGB_ALPHA_EXPAND | RGGL_COLOR_KEY)))
			internalformat = GL_RGBA;
		gluBuild2DMipmaps(
			GL_TEXTURE_2D,
			internalformat,
			bmp.getWidth(),
			bmp.getHeight(),
			outFormat,	// FIXME: Is this correct?
			GL_UNSIGNED_BYTE,
			bmp.getImageData());

	} else {
		glTexImage2D(
			GL_TEXTURE_2D,	// target, must be GL_TEXTURE_2D
			0,			// Base mip map level
			inpFormat,		// Input texture format
			bmp.getWidth(),		// must be 2^k
			bmp.getHeight(),	// must be 2^k
			0,			// No border ("1" gives border)
			outFormat,		// Output texture format
			GL_UNSIGNED_BYTE,	// Data type of pixel data
			bmp.getImageData());
	}

	add_texture_reference( textureID);

	return RGERR_SUCCESS;
}

int rgGLTexture::load_sub_image( rgBMP &bmp, int x, int y, int w, int h,
	int flags)
{
	clear();

	if( w <= 0) return -100;
	if( h <= 0) return -101;
	if( x < 0) return -1;
	if( y < 0) return -2;

	int rstride = (x + w) - bmp.getWidth();
	if( rstride < 0) rstride = 0;
	//if( h > bmp.getHeight()) h = bmp.getHeight();

	const int img_size = w * h * bmp.getColorDepth();
	if( img_size <= 0) return -5;

	rgTempArray <rgBYTE> img( img_size);
	if( img.empty()) return -6;
	bmp.get_sub_image( img.data, x, y, x + w, y + h, rstride);

	width = w;
	height = h;

	/* Get a new texture object from opengl. */
	glGenTextures( 1, &textureID);
	/* If a valid name was not generated then an error occured. The error
	 * might be that there is no current rendering context.
	 */
	if( !textureID) return -7;

	glBindTexture( GL_TEXTURE_2D, textureID);

	/* Select how this texture is to be mapped on to primitives. */
	/* TODO: Support for mip-maps should be added. */
	if( (flags & RGGL_SMOOTH_TEXTURE)) {
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_LINEAR);
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			GL_LINEAR);
	} else {
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_NEAREST);
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			GL_NEAREST);
	}

	/* The created texture will support "wrapping" of its coordinates.
	 * This allows an image to be textured repeatedly across a primitive.
	 */
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	/* Load the image data into OpenGL. */
	GLint inpFormat = 3;	/* RGB */
	GLint outFormat = GL_RGB;
	if( (flags & (RGGL_RGB_ALPHA_EXPAND | RGGL_COLOR_KEY))) { 
		inpFormat = 4;	/* RGBA */
		outFormat = GL_RGBA;
	}

	//cerr << " w = " << w << ", h = " << h << "\n";
	glTexImage2D(
		GL_TEXTURE_2D,	/* target, must be GL_TEXTURE_2D */
		0,		/* Base mip map level*/
		inpFormat,	/* Input texture format */
		w,		/* must be 2^k (not really, fn. will fix) */
		h,		/* must be 2^k (not really, fn. will fix) */
		0,		/* No border (value of 1 would give border) */
		outFormat,	/* Output texture format */
		GL_UNSIGNED_BYTE,/* Data type of pixel data */
		img.data);

	return RGERR_SUCCESS;
}

void rgGLTexture::genAlphaFromRGB( rgBMP &bmp)
{
	const int totalPixels = bmp.getWidth() * bmp.getHeight();
	
	rgBYTE *p = bmp.getImageData();
	int i;
	for( i = 0; i < totalPixels; i++) {

		/* The Alpha component will be the average of the red, green
		 * and blue componenets.
		 */
		p[3] = (p[0] + p[1] + p[2])/3;
		p += 4;	/* Move on to the next pixel */
	}
}

void rgGLTexture::genAlphaFromColorKey( rgBMP &bmp)
{
	const int totalPixels = bmp.getWidth() * bmp.getHeight();

	rgBYTE *p = bmp.getImageData();
	int i;
	for( i = 0; i < totalPixels; i++) {

		/* If this pixel matches with the color key, then it should
		 * have an alpha value of 0 (i.e. is totaly transparent).
		 * Otherwise it has the full alpha value of 255 (i.e. is
		 * totaly opaque).
		 */
		if( RG_RGB( p[0], p[1], p[2]) == colorKey) p[3] = 0;
		else p[3] = 255;

		p += 4;	/* Move on to the next pixel */
	}
}

void loadRGMatrix( const rgMatrix &m)
{
	float d[16];
	d[0] = m.d[0]; d[1] = m.d[4]; d[2] = m.d[8]; d[3] = m.d[12];
	d[4] = m.d[1]; d[5] = m.d[5]; d[6] = m.d[9]; d[7] = m.d[13];
	d[8] = m.d[2]; d[9] = m.d[6]; d[10] = m.d[10]; d[11] = m.d[14];
	d[12] = m.d[3]; d[13] = m.d[7]; d[14] = m.d[11]; d[15] = m.d[15];
	glLoadMatrixf( d);
}

void getRGMatrix( rgMatrix &m)
{
	float d[16];

	int matrixMode;
	glGetIntegerv( GL_MATRIX_MODE, &matrixMode);
	switch( matrixMode) {
	case GL_PROJECTION: glGetFloatv( GL_PROJECTION_MATRIX, d); break;
	case GL_MODELVIEW: glGetFloatv( GL_MODELVIEW_MATRIX, d); break;
	case GL_TEXTURE: glGetFloatv( GL_TEXTURE_MATRIX, d); break;
        // FIXME: GL_COLOR_MATRIX is only available if the GL_ARB_imaging
        //      extension is supported.
        //case GL_COLOR: glGetFloatv( GL_COLOR_MATRIX, d); break;
	default: return;
	}

	m.d[0] = d[0]; m.d[1] = d[4]; m.d[2] = d[8]; m.d[3] = d[12];
	m.d[4] = d[1]; m.d[5] = d[5]; m.d[6] = d[9]; m.d[7] = d[13];
	m.d[8] = d[2]; m.d[9] = d[6]; m.d[10] = d[10]; m.d[11] = d[14];
	m.d[12] = d[3]; m.d[13] = d[7]; m.d[14] = d[11]; m.d[15] = d[15];
}



rgGLLight::rgGLLight()
{
	memset( this, 0, sizeof( rgGLLight));

	id = GL_LIGHT0;
}

void rgGLLight::set_pos( rgFloat x, rgFloat y, rgFloat z, rgFloat w)
{
	pos[0] = x;
	pos[1] = y;
	pos[2] = z;
	pos[3] = w;

	set_pos();
}

void rgGLLight::set_amb( rgFloat r, rgFloat g, rgFloat b, rgFloat a)
{
	amb[0] = r;
	amb[1] = g;
	amb[2] = b;
	amb[3] = a;

	set_amb();
}

void rgGLLight::set_dif( rgFloat r, rgFloat g, rgFloat b, rgFloat a)
{
	dif[0] = r;
	dif[1] = g;
	dif[2] = b;
	dif[3] = a;

	set_dif();
}

void rgGLLight::set_spec( rgFloat r, rgFloat g, rgFloat b, rgFloat a)
{
	spec[0] = r;
	spec[1] = g;
	spec[2] = b;
	spec[3] = a;

	set_spec();
}

rgGLCamera::rgGLCamera()
{
}

rgGLCamera::rgGLCamera( rgVector p, rgVector p_look_at, rgVector up)
{
	cs.p = p;
	look_at( p_look_at, up);
}

void rgGLCamera::look( rgVector eye, rgVector up)
{
	// IMPORTANT NOTE: The camera's "eye" vector must point backwards
	//	since OpenGL uses a postive Z axis as looking away from
	//	the screen. A negative Z axis is look towards (or into) the
	//	screen.
	cs.eye = eye;
	cs.up = up;
	cs.side = calc_side_vec();
}

void rgGLCamera::look_at( rgVector p)
{
	cs.eye = (p - cs.p).normalize();
	const float t = cs.eye.dot( rgVector( 0, 1, 0));
	if( t > 0.8) look_at( p, rgVector( 0, 0, 1));
	else if( t < -0.8) look_at( p, rgVector( 0, 0, -1));
	else look_at( p, rgVector( 0, 1, 0));
}

void rgGLCamera::look_at( rgVector p, rgVector up)
{
	// IMPORTANT NOTE: The camera's "eye" vector must point backwards
	//	since OpenGL uses a postive Z axis as looking away from
	//	the screen. A negative Z axis is look towards (or into) the
	//	screen.
	cs.eye = (p - cs.p).normalize();
	cs.up = (up - (cs.eye.dot( up) * cs.eye)).normalize();
	cs.side = calc_side_vec();
}

void rgGLCamera::apply() const
{
	// Convert the coordinate system into matrix form
	rgCoordSystem tmp( cs);
	tmp.rotate( RG_DEGTORAD( 180.0f), cs.up);
	rgMatrix m( tmp);
	// For a camera, the translation must be the reverse of the position.
	m.d[3] *= -1.0f;
	m.d[7] *= -1.0f;
	m.d[11] *= -1.0f;
	loadRGMatrix( m);
}

rgVector rggl_unproject( float x, float y, float z)
{
	GLdouble model[16];
	GLdouble proj[16];
	GLint view[4];
	GLdouble obj[3];
	glGetDoublev( GL_MODELVIEW_MATRIX, model);
	glGetDoublev( GL_PROJECTION_MATRIX, proj);
	glGetIntegerv( GL_VIEWPORT, view);
	if( y < 0) y = 0;
	if( y >= view[3]) y = view[3] - 1;
	// FIXME: Should check if gluUnProject returns an error.
	gluUnProject(
		x, view[3] - y, z,
		model,
		proj,
		view,
		obj, obj + 1, obj + 2);

	return rgVector( obj[0], obj[1], obj[2]);
}

/*
void rggl_pick_ray( int x, int y, float fov, rgVector &rp, rgVector &rd)
{
	GLint v[4];
	GLint &width = v[2];
	GLint &height = v[3];
	glGetIntegerv( GL_VIEWPORT, v);

	if( x < 0) x = 0;
	if( x >= width) x = width - 1;
	if( y < 0) y = 0;
	if( y >= height) y = height - 1;
	//width >>= 1;
	//height >>= 1;

	const int w = width >> 1;
	const int h = height >> 1;
	x -= w;
	y -= h;
	const float hyp = float( width) / sin( RG_DEGTORAD( fov * 0.5f));
	const float dis = sqrt( hyp * hyp - float( width * width));

	rgMatrix m;
	getRGMatrix( m);
	m.invert();

	rp = m.pointMul( rgVector( x, -y, 0));
	//const rgVector xray( x, 0, -dis);
	//const rgVector yray( 0, -y, -dis);
	//const rgVector yray;
	rd = m * (rgVector( x, -y, -dis).normalize());

	//rp = m.pointMul( rgVector( 0, 0, dis));
	//const rgVector xray( x, 0, -dis);
	//const rgVector yray( 0, -y, -dis);
	//const rgVector yray;
	//rd = m * ((xray + yray).normalize());

	//std::cerr << "dis = " << dis << "\n";
	//std::cerr << "rp = " << rp << "\n";
	//std::cerr << "rd = " << rd << "\n";
	//std::cerr << "x = " << x << ", y = " << y << "\n";
}
*/

/*
void rggl_unproject( rgVector *v, const int count)
{
	if( !v) return;
	if( count <= 0) return;

	// Remember current matrix mode.
        GLint mm;
        glGetIntegerv( GL_MATRIX_MODE, &mm);

        if( GL_PROJECTION != mm) glMatrixMode( GL_PROJECTION);
	GLdouble pj[16];
	glGetDoublev( GL_PROJECTION_MATRIX, pj);

        glMatrixMode( GL_MODELVIEW);
	GLdouble mv[16];
	glGetDoublev( GL_MODELVIEW_MATRIX, mv);

	GLint vp[4];
	glGetIntegerv( GL_VIEWPORT, vp);

	double dx, dy, dz;
	for( int i = 0; i < count; i++) {
		gluUnProject( v[i].x, v[i].y, v[i].z, mv, pj, vp,
			&dx, &dy, &dz);
		v[i].x = dx;
		v[i].y = dy;
		v[i].z = dz;
	}
	
	// Restore original matrix mode
	if( GL_MODELVIEW != mm) glMatrixMode( mm);
}
*/

void rggl_get_ortho_dimensions( float &w, float &h, rgMatrix &m)
{
	GLint mm;
	glGetIntegerv( GL_MATRIX_MODE, &mm);
	if( GL_PROJECTION != mm) glMatrixMode( GL_PROJECTION);
	getRGMatrix( m);
	if( GL_PROJECTION != mm) glMatrixMode( mm);
	w = 2.0f / m.d[0];
	h = 2.0f / m.d[5];
}

inline void hlp_window_to_xy_world( const float p_x, const float p_y,
	const float inv_w_ow, const float inv_h_oh, const float ow,
	const float oh, rgVector &u)
{
        u.x = (p_x * inv_w_ow) - (ow * 0.5f);
        /* NOTE: Y axis is inverted for window coordinates (p_y). */
        u.y = -((p_y * inv_h_oh) - (oh * 0.5f));
        u.z = 0.0f;
}

void rggl_ortho_window_to_world( const float w, const float h, rgVector *v,
	const int count, const rgMatrix *inv)
{
	if( !v) return;
	if( count <= 0) return;
        if( (0.0f == w) || (0.0f == h)) return;

	float ow, oh;
	rgMatrix m;
        rggl_get_ortho_dimensions( ow, oh, m);

	const float inv_w_ow = ow / w;
	const float inv_h_oh = oh / h;
	int i;
	for( i = 0; i < count; i++) {
	        hlp_window_to_xy_world( v[i].x, v[i].y, inv_w_ow, inv_h_oh,
			ow, oh, v[i]);
	}
	// Apply the inverse transformation (if an inverse matrix was given).
	if( inv) for( i = 0; i < count; i++) v[i] = (*inv).pointMul( v[i]);
}

void rggl_ortho_world_to_window( const float w, const float h, rgVector *v,
	int count)
{
	rgMatrix m;
	float ow, oh;
        rggl_get_ortho_dimensions( ow, oh, m);

	const float inv_w_ow = 0.5f * w;
	const float inv_h_oh = -(0.5f * h);
	while( (count--) > 0) {
		*v = m.pointMul( *v);
		v->x = (v->x + 1.0f) * inv_w_ow;
		v->y = (v->y + 1.0f) * inv_h_oh;
		// XXX: The Z component is not used (ignored).
		v++;
	}
}

/*
void window_to_inverse_world( int p_x, int p_y, GLView &g, rgVector &u)
{
        window_to_xy_world( p_x, p_y, g, u);
        // Apply effects of current modelview matrix.
        rgMatrix m;
        getRGMatrix( m);
        m.invert();
        u = m.pointMul( u);
}
*/

rgGLViewMatrix::rgGLViewMatrix()
{
	trans_order = RGGL_VIEW_MATRIX_SRT;
}

void rgGLViewMatrix::set_transformation_order( int order)
{
	switch( order) {
	case RGGL_VIEW_MATRIX_SRT:
	case RGGL_VIEW_MATRIX_STR:
	case RGGL_VIEW_MATRIX_RTS:
	case RGGL_VIEW_MATRIX_TRS:
		trans_order = order;
		reset();
		break;
	default: return;
	}
};

void rgGLViewMatrix::translate( rgVector v)
{
	rgMatrix m;
	m.translate( v);
	camT = m * camT;

	m.loadIdentity();
	m.translate( -1.0f * v);
	icamT = icamT * m;
}

void rgGLViewMatrix::translate( rgFloat x, rgFloat y, rgFloat z)
{
	translate( rgVector( x, y, z));
}

void rgGLViewMatrix::rotate( const rgFloat ang, rgVector axis)
{
	rgMatrix m;
	m.rotate( ang, axis);
	camR = m * camR;

	m.loadIdentity();
	m.rotate( -ang, axis);
	icamR = icamR * m;
}

void rgGLViewMatrix::scale( rgFloat x, rgFloat y, rgFloat z)
{
	// Scaling by zero is not allowed for any dimension. This is because
	// it would cause a divide by zero when determining the inverse
	// scaling matrix.
	if( (0.0f == x) || (0.0f == y) || (0.0f == z)) return;

	rgMatrix m;
	m.scale( x, y, z);
	camS = m * camS;

	m.loadIdentity();
	m.scale( 1.0f / x, 1.0f / y, 1.0f / z);
	icamS = icamS * m;
}

void rgGLViewMatrix::reset()
{
	// Reset all view matricies.
	camM.loadIdentity();
	camT = camR = camS = camM;

	// Reset all inverse matricies.
	icamM = icamT = icamR = icamS = camM;
}

void rgGLViewMatrix::update()
{
	switch( trans_order) {
	case RGGL_VIEW_MATRIX_SRT:
		camM = (camT * camR) * camS;
		icamM = (icamS * icamR) * icamT;
		break;

	case RGGL_VIEW_MATRIX_STR:
		camM = (camR * camT) * camS;
		icamM = (icamS * icamT) * icamR;
		break;

	case RGGL_VIEW_MATRIX_RTS:
		camM = (camS * camT) * camR;
		icamM = (icamR * icamT) * icamS;
		break;

	case RGGL_VIEW_MATRIX_TRS:
		camM = (camS * camR) * camT;
		icamM = (icamT * icamR) * icamS;
		break;
	}
}

void rgGLViewMatrix::update( bool tm, bool rm, bool sm)
{
	rgMatrix *a = 0, *b = 0, *c = 0;
	rgMatrix *ia = 0, *ib = 0, *ic = 0;
	switch( trans_order) {
	case RGGL_VIEW_MATRIX_SRT:
		if( tm) { a = &camT; ia = &icamT;}
		if( rm) { b = &camR; ib = &icamR;}
		if( sm) { c = &camS; ic = &icamS;}
		break;

	case RGGL_VIEW_MATRIX_STR:
		if( rm) { a = &camR; ia = &icamR;}
		if( tm) { b = &camT; ib = &icamT;}
		if( sm) { c = &camS; ic = &icamS;}
		break;

	case RGGL_VIEW_MATRIX_RTS:
		if( sm) { a = &camS; ia = &icamS;}
		if( tm) { b = &camT; ib = &icamT;}
		if( rm) { c = &camR; ic = &icamR;}
		break;

	case RGGL_VIEW_MATRIX_TRS:
		if( sm) { a = &camS; ia = &icamS;}
		if( rm) { b = &camR; ib = &icamR;}
		if( tm) { c = &camT; ic = &icamT;}
		break;
	}

	if( a) {
		camM = *a; icamM = *ia;
		if( b) { camM = camM * (*b); icamM = (*ib) * icamM;}
		if( c) { camM = camM * (*c); icamM = (*ic) * icamM;}

	} else {

		if( b) {
			camM = *b; icamM = *ib;
			if( c) { camM = camM * (*c); icamM = (*ic) * icamM;}
		} else {
			if( c) { camM = *c; icamM = *ic;}
		}
	}
}

void rggl_print_error()
{
	rggl_print_error( std::cerr);
}

void rggl_print_error( std::ostream &o)
{
	GLenum err;
	switch( (err = glGetError())) {
	case GL_NO_ERROR: o << "GL_NO_ERROR"; break;
	case GL_INVALID_ENUM: o << "GL_INVALID_ENUM"; break;
	case GL_INVALID_VALUE: o << "GL_INVALID_VALUE"; break;
	case GL_INVALID_OPERATION: o << "GL_INVALID_OPERATION"; break;
	case GL_STACK_OVERFLOW: o << "GL_STACK_OVERFLOW"; break;
	case GL_STACK_UNDERFLOW: o << "GL_STACK_UNDERFLOW"; break;
	case GL_OUT_OF_MEMORY: o << "GL_OUT_OF_MEMORY"; break;
#ifndef WINDOWS_OS
	case GL_TABLE_TOO_LARGE: o << "GL_TABLE_TOO_LARGE"; break;
#endif
	default: o << "Unrecognized GL Error (" << err << ")"; break;
	}
}

void rgGLViewMatrix::do_view_translate( const float width, const float height,
	const rgVector src, const rgVector des, rgMatrix *m, rgMatrix *im,
	const bool set)
{
	rgVector v[2] = { src, des};
	rgVector &a = v[0];
	rgVector &b = v[1];
	
	rggl_ortho_window_to_world( width, height, v, 2, &icamM);

	if( m) m->translate( b - a);
	if( im) im->translate( a - b);

	if( set) {
		translate( b - a);
		update();

	} else {
		rgMatrix tmp_camT( camT);
		rgMatrix tmp_icamT( icamT);
		translate( b - a);
		update();
		camT = tmp_camT;
		icamT = tmp_icamT;
	}
}

void rgGLViewMatrix::do_view_3way_rotate( const float width,
	const float height, const rgVector src,
	const rgVector des, rgMatrix *m, rgMatrix *im, const bool set)
{
	const int x = int( int( 3 * src.x) / width);
	const int y = int( int( 3 * src.y) / height);
	const int id = y * 3 + x;
	if( 4 == id) return;
	do_view_4way( width, height, src, des, m, im, set);
}

void rgGLViewMatrix::do_view_4way( const float width,
	const float height, const rgVector src,
	const rgVector des, rgMatrix *m, rgMatrix *im, const bool set)
{
	const int x = int( int( 3 * src.x) / width);
	const int y = int( int( 3 * src.y) / height);
	const int id = y * 3 + x;
	switch( id) {
	case 4:
		do_view_translate( width, height, src, des, m, im, set);
		break;

	case 0:
	case 2:
	case 6:
	case 8:
		do_view_circular_rotate( width, height, src, des,
			rgVector( 0,0,-1), m, im, set);
		break;

	case 1:
	case 7:
		do_view_rotate( rgVector( 0, height, 0), src, des,
			rgVector( 1, 0, 0), m, im, set);
		break;

	case 3:
	case 5:
		do_view_rotate( rgVector( width, 0, 0), src, des,
			rgVector( 0, 1, 0), m, im, set);
		break;
	}
}

void rgGLViewMatrix::do_view_rotate( const rgVector forward,
	const rgVector src, const rgVector des, const rgVector axis,
	rgMatrix *m, rgMatrix *im, const bool set)
{
	const rgVector w( des - src);
	const float inv_len = 1.0f / forward.length();
	const float ang = (w.dot( inv_len * forward) * inv_len)
		* RG_180_DEG_IN_RAD;
	const rgMatrix tmp_camR( camR);
	const rgMatrix tmp_icamR( icamR);
	rotate( ang, axis);
	if( m) m->rotate( ang, axis);
	if( im) m->rotate( -ang, axis);
	update();
	if( set) return;
	camR = tmp_camR;
	icamR = tmp_icamR;
}

void rgGLViewMatrix::do_view_circular_rotate( const float width,
	const float height, const rgVector src, const rgVector des,
	const rgVector axis, rgMatrix *m, rgMatrix *im, const bool set)
{
	const float cenX = width * 0.5f;
	const float cenY = height * 0.5f;
	const rgVector u( src.x - cenX, src.y - cenY, 0.0f);
	const rgVector v( des.x - cenX, des.y - cenY, 0.0f);

	float ang = rgDotAngle( u, v);
	if( (u.cross( v)).z < 0.0f) ang = -ang;

	const rgMatrix tmp_camR( camR);
	const rgMatrix tmp_icamR( icamR);
	rotate( ang, axis);
	if( m) m->rotate( ang, axis);
	if( im) im->rotate( -ang, axis);
	update();
	if( set) return;
	camR = tmp_camR;
	icamR = tmp_icamR;
}

void rgGLViewMatrix::do_view_scale( const float height, const rgVector src,
	rgVector des, const float k, rgMatrix *m, rgMatrix *im,
	const bool set)
{
	rgVector w = des - src;
	const float y = pow( 2.0,
		k *(w.dot( rgVector( 0, 1, 0)) / height));
	
	if( m) m->scale( y, y, y);
	if( im) im->scale( 1.0/y, 1.0/y, 1.0/y);

	if( set) {
		scale( y, y, y);
		update();

	} else {
		rgMatrix tmp_camS( camS);
		rgMatrix tmp_icamS( icamS);
		scale( y, y, y);
		update();
		camS = tmp_camS;
		icamS = tmp_icamS;
	}
}


int rggl_read_pixels( rgBMP &bmp, const GLint buf, const int x, const int y,
	const int width, const int height)
{
	// FIXME: Are "x" and "y" allowed to be negative?
	//if( x < 0) return -1;
	//if( y < 0) return -2;
	if( width <= 0) return -3;
	if( height <= 0) return -4;
	if( bmp.generate_bmp( width, height, 24)) return -5;
	glReadPixels( x, y, width, height, GL_RGB, GL_UNSIGNED_BYTE,
		bmp.getImageData());

/*
	rgBYTE *p = bmp.getImageData();
	{
		for( int i = 0; i < width; i++) {
			p[i*3 + 0] = 0xFF;
			p[i*3 + 1] = 0x00;
			p[i*3 + 2] = 0x00;

			p[i*3 + (height - 1) * (width * 3) + 0] = 0x00;
			p[i*3 + (height - 1) * (width * 3) + 1] = 0xFF;
			p[i*3 + (height - 1) * (width * 3) + 2] = 0x00;
		}
	}

	{
		for( int i = 1; i < (height - 1); i++) {
			p[i*(width * 3) + 0] = 0x00;
			p[i*(width * 3) + 1] = 0x00;
			p[i*(width * 3) + 2] = 0xFF;

			p[i*(width * 3) + ((width - 1) * 3) + 0] = 0xFF;
			p[i*(width * 3) + ((width - 1) * 3) + 1] = 0x00;
			p[i*(width * 3) + ((width - 1) * 3) + 2] = 0xFF;
		}
	}
*/
	return 0;
}
