/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "ai.h"
#include "engine/engine_game.h"

std::map <const std::string, int (*)( const gmCCArg &)> phAI::m_ccmd;

extern "C" gmThing *spw_phAI()
{
	return new phAI();
}

extern "C" int del_phAI( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phAI)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phAI", spw_phAI);

	p_ge->m_thing_manager->add_to_delete_list( "phAI", del_phAI);

	return 0;
}

phAI::phAI()
{
	m_host = 0;

	if( !(m_ccmd.empty())) return;
	add_console_command( "set_host", phAI_set_host, m_ccmd);
}

phAI::~phAI()
{
}

int phAI::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.find( key);
	if( iter == m_ccmd.end()) return gmThing::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phAI::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); iter++)
		tmp += p_prefix + phAI::type() + "::" + (iter->first) + "\n";
	tmp += gmThing::list_ccmd( p_prefix);
	return tmp;
}

int phAI_set_host( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: set_host <host>\n"
			"NOTE: Some AI's will only accept specific hosts.\n");
		return GM_THING_ERR_PARAM;
	}

	ThingControlBlock *t = (p.ge->m_thing_manager)->get( args[0]);
	if( !t) {
		p.ge->rcout( "No such thing \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}

	phAI *ai = (phAI *)(p.thing);
	if( !(ai->set_host( t))) {
		p.ge->rcout( "\"" + args[0] + "\" is an unacceptable host for "
			"AI of type \"" + ai->type() + "\"\n");
		return GM_THING_ERR_GENERIC;
	}

        return 0;
}
