/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "team.h"
#include "player.h"
#include "coach.h"
#include "engine/engine_game.h"
#include <rg3DRom.h>

std::map <const std::string, int (*)( const gmCCArg &)> phTeam::m_ccmd;

extern "C" gmThing *spw_phTeam()
{
	return new phTeam();
}

extern "C" int del_phTeam( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phTeam)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phTeam", spw_phTeam);

	p_ge->m_thing_manager->add_to_delete_list( "phTeam", del_phTeam);

	return 0;
}

phTeam::phTeam()
{
	m_head_coach = 0;

	if( !(m_ccmd.empty())) return;
	add_console_command( "add_person", phTeam_add_person, m_ccmd);
	add_console_command( "add_person_list", phTeam_add_person_list,
		m_ccmd);
}

phTeam::~phTeam()
{
}

int phTeam::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.find( key);
	if( iter == m_ccmd.end()) return gmThing::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phTeam::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); iter++)
		tmp += p_prefix + phTeam::type() + "::" + (iter->first) + "\n";
	tmp += gmThing::list_ccmd( p_prefix);
	return tmp;
}

int phTeam_add_person( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 2) {
                p.ge->rcout( "usage: add_person <destination list> <person>\n"
			"NOTE: Person must be of type phCoach or phPlayer.\n"
			"So for example, you cannot add officials ("
				"phReferee).\n");
                return GM_THING_ERR_PARAM;
        }
	phTeam *t = (phTeam *)(p.thing);
	const std::string emsg( t->add_person( args[0], args[1]));
	if( emsg.length() > 0) {
		p.ge->rcout( emsg);
		return GM_THING_ERR_GENERIC;
	}
        return 0;
}

int phTeam_add_person_list( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 1) {
                p.ge->rcout( "usage: add_person_list <list name>\n"
			"NOTE: The list name \"head_coach\" is reserved and "
				"may not be used.\n");
                return GM_THING_ERR_PARAM;
        }
	phTeam *t = (phTeam *)(p.thing);
	const std::string emsg( t->create_person_list( args[0]));
	if( emsg.length() > 0) {
		p.ge->rcout( emsg);
		return GM_THING_ERR_GENERIC;
	}
        return 0;
}


ThingControlBlock *phTeam::get_person( const std::string &p_list,
	const std::string &p_person)
{
	std::map <const std::string, ThingControlBlock *> *tmp =
		get_person_list( p_list);
	if( !tmp) return 0;

	std::map <const std::string, ThingControlBlock *> :: const_iterator
		iter = tmp->find( p_person);
	if( iter == tmp->end()) return 0;
	return (iter->second);
}

std::map <const std::string, ThingControlBlock *> *phTeam::get_person_list(
	const std::string &key)
{
	std::map <const std::string,
		std::map <const std::string, ThingControlBlock *> *> ::
		const_iterator iter = m_person.find( key);
	if( iter == m_person.end()) return 0;
	return (iter->second);
}

std::string phTeam::add_person( const std::string &p_list,
	const std::string &p_person)
{
	ThingControlBlock *t = (ge())->m_thing_manager->get( p_person);
	if( !t) return ("No such thing \"" + p_person + "\"");
	if( t->is_free())
		return ("Thing \"" + p_person + "\" is free (unusable)");

	if( "head_coach" == p_list) {
		if( m_head_coach)
			return "Must fire current head coach first";
		if( !(t->m_thing->is_a( "phCoach")))
			return "\"" + p_person + "\" is not a coach";
		m_head_coach = t;
		((phCoach *)(t->m_thing))->set_team( tcb());
		return std::string();
	}

	if( !(t->m_thing->is_a( "phPerson")))
		return "\"" + p_person + "\" is not a person";

	std::map <const std::string, ThingControlBlock *> *tmp =
		get_person_list( p_list);
	if( !tmp) return ("No such person list \"" + p_list + "\"");

	std::string emsg( add_person( tmp, t));
	if( emsg.length() > 0) emsg = "List \"" + p_list + "\": " + emsg;
	else ((phPerson *)(t->m_thing))->set_team( tcb());

	return emsg;
}

std::string phTeam::add_person(
	std::map <const std::string, ThingControlBlock *> *p_list,
	ThingControlBlock *p_tcb)
{
	typedef std::pair < const std::string, ThingControlBlock *> STPair;
	if( false == (p_list->insert( STPair( p_tcb->m_key, p_tcb))).second)
		return "Could not add thing \"" + p_tcb->m_key + "\"";
	return std::string();
}

std::string phTeam::remove_person( const std::string &p_list,
	const std::string &p_person)
{
	// Get the named list.
	std::map <const std::string, ThingControlBlock *> *tmp =
		get_person_list( p_list);
	if( !tmp) return ("No such person list \"" + p_list + "\"");

	// Find them in the given list.
	std::map <const std::string, ThingControlBlock *> :: iterator iter =
		tmp->find( p_person);
	if( iter == tmp->end()) return ("Person \"" + p_person + "\" not in \""
		+ p_list + "\" list");

	tmp->erase( iter);
	return std::string();
}

std::string phTeam::create_person_list( const std::string &key)
{
	if( "head_coach" == key) return "list \"" + key + "\" is reserved";
	typedef std::pair <const std::string,
		std::map <const std::string, ThingControlBlock *> *> SMPair;
	std::map <const std::string, ThingControlBlock *> *tmp =
		new std::map <const std::string, ThingControlBlock *>;
	if( !tmp) return "Could not create new \"map <string, tcb>\"";
	if( false == (m_person.insert( SMPair( key, tmp))).second) {
		delete tmp;
		return "Could not add person list \"" + key + "\", "
			"(possibly a duplicate?)";
	}
	return std::string();
}

std::string phTeam::remove_person_list( const std::string &key)
{
	if( "head_coach" == key)
		return "list \"" + key + "\" cannot be removed";
	std::map <const std::string,
		std::map <const std::string, ThingControlBlock *> *> ::
		iterator iter = m_person.find( key);
	if( iter == m_person.end()) return "No such list \"" + key + "\"";
	delete (iter->second);
	m_person.erase( iter);
	return std::string();
}

